/*-------------------------------------------------------------------------+
| Copyright 2021 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.storage;

import static org.apache.commons.io.FilenameUtils.getExtension;
import static org.eclipse.core.resources.ResourcesPlugin.getWorkspace;
import static org.eclipse.emf.common.util.URI.createFileURI;
import static org.fortiss.tooling.ext.reuse.ToolingReuseActivator.getDefault;
import static org.fortiss.tooling.kernel.utils.EMFResourceUtils.createNewEclipseWorkspaceProject;
import static org.fortiss.tooling.kernel.utils.EMFResourceUtils.loadModelFromFile;
import static org.fortiss.tooling.kernel.utils.EMFResourceUtils.refreshWorkspace;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.showError;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.common.util.URI;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;

/**
 * Manages the separate (local) storage of {@link ReuseLibrary}s.
 *
 * @author bergemann
 */
public class ReuseLibraryStorageManager {

	/** The name of the general reuse (library) directory. */
	public static final String AF3_REUSE_DIRECTORY_NAME = "AF3-Reuse-Directory";

	/** AF3 reuse library files extension */
	public static final String AF3_REUSE_LIBRARY_FILES_EXTENSION = "af3_rlib";

	/** The current workspace root. */
	public static final IWorkspaceRoot WORKSPACE_ROOT = getWorkspace().getRoot();

	/** The File handler of the reuse project. */
	public static final File REUSE_PROJECT_DIR = new File(
			WORKSPACE_ROOT.getLocation() + File.separator + AF3_REUSE_DIRECTORY_NAME);

	/** Returns whether the given file is a {@link ReuseLibrary} file. */
	private static boolean isLibraryFile(File file) {
		return getExtension(file.getName()).equals(AF3_REUSE_LIBRARY_FILES_EXTENSION);
	}

	/**
	 * Returns the correct name for the local folder of the given
	 * {@link ReuseLibrary}.
	 */
	private static String getLibraryFolderName(ReuseLibrary library) {
		// folder gets also UUID, because libraries could have identical names across
		// different projects (UUID contains never a minus symbol)
		return library.getName() + "-" + library.getLibraryUUID();
	}

	/**
	 * Returns the correct name for the local file of the given
	 * {@link ReuseLibrary}.
	 */
	private static String getLibraryFileName(ReuseLibrary library) {
		// inside the library folder, there is currently only one file with all the af3
		// stuff (for this library)
		return library.getName() + "." + AF3_REUSE_LIBRARY_FILES_EXTENSION;
	}

	/** Returns the {link IFolder} of the given {@link ReuseLibrary}. */
	private static IFolder getLibraryFolder(ReuseLibrary library) {
		String libFolderName = getLibraryFolderName(library);
		IProject project = WORKSPACE_ROOT.getProject(AF3_REUSE_DIRECTORY_NAME);
		return project.getFolder(libFolderName);
	}

	/** Returns the {link IFile} of the given {@link ReuseLibrary}. */
	private static IFile getLibraryFile(ReuseLibrary library) {
		String libraryFolderName = getLibraryFolderName(library);
		String af3LibraryFileName = getLibraryFileName(library);
		return WORKSPACE_ROOT.getFile(new Path(
				AF3_REUSE_DIRECTORY_NAME + File.separator + libraryFolderName + File.separator + af3LibraryFileName));
	}

	/**
	 * Returns all stored {@link ReuseLibrary}s inside the reuse directory.
	 *
	 * @return List of all stored reuse libraries
	 */
	public static List<ReuseLibrary> getAllSeparatelyStoredReuseLibraries() {
		List<ReuseLibrary> libraries = new ArrayList<ReuseLibrary>();
		if (REUSE_PROJECT_DIR.exists() && REUSE_PROJECT_DIR.isDirectory()) {
			File[] filesInDirectory = REUSE_PROJECT_DIR.listFiles();
			for (File directoryFile : filesInDirectory) {
				if (directoryFile.isDirectory()) {
					for (File libraryFile : directoryFile.listFiles()) {
						if (libraryFile.isFile() && isLibraryFile(libraryFile)) {
							URI sourceFileURI = createFileURI(libraryFile.getAbsolutePath());
							ReuseLibrary library = (ReuseLibrary) loadModelFromFile(sourceFileURI);
							libraries.add(library);
						}
					}
				}
			}
		}
		return libraries;
	}

	/**
	 * Saves a {@link ReuseLibrary}. Will override file if a library with identical
	 * name and UUID already exists.
	 *
	 * @param library The target reuse library
	 * @return boolean Whether the save was successful (true) or not (false)
	 * @throws IOException
	 * @throws CoreException
	 */
	public static boolean saveReuseLibrary(ReuseLibrary library) throws IOException, CoreException {
		return saveReuseLibrary(library, true);
	}

	/**
	 * Saves a {@link ReuseLibrary}. It can be specified if a file should be
	 * overridden if a library with identical name and UUID already exists.
	 *
	 * @param library  The target reuse library
	 * @param override The flag if possible existing file should be overridden
	 *                 (true) or not (false)
	 * @return boolean Whether the save was successful (true) or not (false)
	 * @throws IOException
	 * @throws CoreException
	 */
	public static boolean saveReuseLibrary(ReuseLibrary library, Boolean override) throws IOException, CoreException {
		ensureReuseProjectExists();

		IFolder folder = getLibraryFolder(library);
		if (!folder.exists()) {
			folder.create(IResource.NONE, true, null);
		}

		IFile file = getLibraryFile(library);
		if (override || !file.exists()) {
			createNewEclipseWorkspaceProject(file, library);
			return true;
		}
		return false;
	}

	/**
	 * Makes sure that a general reuse project exists and is added to the current
	 * workspace.
	 */
	private static void ensureReuseProjectExists() {
		// if project directory exists, do noting
		String absoluteProjectPath = REUSE_PROJECT_DIR.getAbsolutePath();
		if (new File(absoluteProjectPath).isDirectory()) {
			return;
		}

		IProject project = WORKSPACE_ROOT.getProject(AF3_REUSE_DIRECTORY_NAME);
		try {
			// (re)create project with new directory
			// (if only the directory was deleted, delete internal project before
			// recreating it)
			if (project.exists()) {
				project.delete(false, new NullProgressMonitor());
			}
			project.create(null);
			project.open(null);
			refreshWorkspace();
		} catch (CoreException e) {
			error(getDefault(), "Error during creating new project", e);
			showError("Cannot create reuse project!");
		}
	}

	/**
	 * Deletes the local folder of the given {@link ReuseLibrary}.
	 *
	 * @param library The target reuse library
	 * @throws CoreException
	 */
	public static void deleteReuseLibraryFolder(ReuseLibrary library) throws CoreException {
		IFolder folder = getLibraryFolder(library);
		folder.delete(true, new NullProgressMonitor());
	}

}
